/*************************************************************/
/* iTraceroute (part of iSPY) Project                        */
/* Copyright (c) 2008, Zheng Zhang, Purdue University        */
/*************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>

#include <iostream>
#include <sstream>

#include "NetUtil.h"


string NetUtil::ip_uint32_to_str(uint32 addr)
{
    struct sockaddr_in mysocket;

    mysocket.sin_addr.s_addr = addr;
    return string (inet_ntoa(mysocket.sin_addr));
}


string NetUtil::print_ip(struct _iphdr *p)
{
    ostringstream oss (ostringstream::out);

    struct sockaddr_in mysocket;

    mysocket.sin_addr.s_addr = p->saddr;
    oss << "IP " << inet_ntoa(mysocket.sin_addr);
    mysocket.sin_addr.s_addr = p->daddr;
    oss << " > " << inet_ntoa(mysocket.sin_addr);
  
    switch(p->protocol){
    case IPPROTO_ICMP:
        oss << " ICMP";
        break;
    case IPPROTO_TCP:
        oss << " TCP";
        break;
    case IPPROTO_UDP:
        oss << " UDP" ;
        break;
    default:
        oss << " Proto=" << p->protocol;
        break;
    }
    return oss.str();
}


string NetUtil::print_tcp(struct _tcphdr *p)
{
    ostringstream oss (ostringstream::out);    
    
    oss << "TCP "<< ntohs(p->dest)
        << " > " << ntohs(p->source) << " ";
   
    if(p->fin)
        oss << "FIN.";
    if(p->syn)
        oss << "SYN.";
    if(p->rst)
        oss << "RST.";
    if(p->psh)
        oss << "PUSH.";
    if(p->ack)
        oss << "ACK.";
    if(p->urg)
        oss << "URG.";

    oss << " ";
    uint32 seq = ntohl(p->seq);
    uint32 ack = ntohl(p->ack_seq);
    oss << "seq=" << seq << " ack=" << ack;
    return oss.str();
}


string NetUtil::print_icmp(struct _icmphdr *p)
{
    static char* text[6] = {
        "0 (net unreachable)",
        "1 (host unreachable)",
        "2 (protocol unreachable)",
        "3 (port unreachable)",
        "4 (fragmentation needed and DF set)",
        "5 (source route failed)"
    };
    
/*
    0 = net unreachable;
    1 = host unreachable;
    2 = protocol unreachable;
    3 = port unreachable;
    4 = fragmentation needed and DF set;
    5 = source route failed.
*/

    ostringstream oss (ostringstream::out);    
    
    if (p->code < 6)  //p->code is unsigned, so > 0
        oss << "ICMP code = " << text[p->code];
    else
        oss << "ICMP code = " << p->code << " (unknown)";

    return oss.str();
}


/*
 * in_cksum2 --
 *      Checksum routine for Internet Protocol family headers (C Version)
 */
unsigned short NetUtil::in_cksum2(unsigned short *addr, int len)
{
    register int sum = 0;
    u_short answer = 0;
    register u_short *w = addr;
    register int nleft = len;
  
    /*
     * Our algorithm is simple, using a 32 bit accumulator (sum), we add
     * sequential 16 bit words to it, and at the end, fold back all the
     * carry bits from the top 16 bits into the lower 16 bits.
     */
    while (nleft > 1)  {
        sum += *w++;
        nleft -= 2;
    }
  
    /* mop up an odd byte, if necessary */
    if (nleft == 1) {
        *(u_char *)(&answer) = *(u_char *)w ;
        sum += answer;
    }
  
    /* add back carry outs from top 16 bits to low 16 bits */
    sum = (sum >> 16) + (sum & 0xffff);     /* add hi 16 to low 16 */
    sum += (sum >> 16);                     /* add carry */
    answer = ~sum;                          /* truncate to 16 bits */
    return(answer);
}


void NetUtil::ip_gen(char *packet, unsigned char protocol, uint32 saddr, uint32 daddr, 
    unsigned short ipid, unsigned short length, uint8 ttl)
{
    struct _iphdr *iphdr;
  
    iphdr = (struct _iphdr *)packet;
    memset((char *)iphdr, '\0', sizeof(struct _iphdr));
  
    iphdr->ihl = 5;
    iphdr->version = IPVERSION;
    
#ifdef IP_LEN_HORDER
    iphdr->tot_len = length;
#else
    iphdr->tot_len = htons(length);
#endif /* IP_LEN_HORDER */
  
    iphdr->id = htons(ipid);
    iphdr->ttl = ttl;
    iphdr->protocol = protocol;
    
    iphdr->saddr = saddr;
    iphdr->daddr = daddr;
    
    iphdr->check = (unsigned short)in_cksum2((unsigned short *)iphdr, sizeof(struct _iphdr));
    return;
}


void NetUtil::tcp_gen(char *packet, uint16 sport, uint16 dport, 
                      uint32 seq, uint32 ack, char f_syn, char f_ack, char f_rst)
{
    struct _tcphdr *tcp;
  
    tcp = (struct _tcphdr *)packet;
    memset((char *)tcp, '\0', sizeof(struct _tcphdr));
  
    tcp->source = htons(sport);
    tcp->dest = htons(dport);
  
    tcp->seq = htonl(seq);
    tcp->ack_seq = htonl(ack);
  
    tcp->res1 = 0;
    tcp->doff = TH_OFFSET;

    tcp->window = htons(TCP_WINDOW_SIZE);
    tcp->syn = f_syn;
    tcp->ack = f_ack;
    tcp->rst = f_rst;
    return;
}


unsigned short 
NetUtil::trans_check(unsigned char proto, char *packet, int length, 
                     uint32 saddr, uint32 daddr)
{
    struct psuedohdr  {
        struct in_addr source_address;
        struct in_addr dest_address;
        unsigned char place_holder;
        unsigned char protocol;
        unsigned short length;
    } psuedohdr;
    
    char *psuedo_packet;
    unsigned short answer;
    
    psuedohdr.protocol = proto;
    psuedohdr.length = htons(length);
    psuedohdr.place_holder = 0;
    
    psuedohdr.source_address.s_addr = saddr;
    psuedohdr.dest_address.s_addr = daddr;
    
    if((psuedo_packet = (char *)malloc(sizeof(psuedohdr) + length)) == NULL)  {
        throw MyException(str_log(ERROR, "malloc"));
    }
    
    memcpy(psuedo_packet, &psuedohdr,sizeof(psuedohdr));
    memcpy((psuedo_packet + sizeof(psuedohdr)),
           packet,length);
    
    answer = (unsigned short)in_cksum2((unsigned short *)psuedo_packet, (length + sizeof(psuedohdr)));
    free (psuedo_packet);
    return answer;
}


uint32 
NetUtil::getRoute (uint32 destip) 
{
  int sock;
  struct sockaddr_in addr, name;
  
  if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
      log (WARN, "socket(), ret = -1. %s\n", strerror(errno));
      return 0;
  }
  
  bzero(&addr, sizeof(struct sockaddr_in));

  addr.sin_port = htons(1234);
  addr.sin_addr.s_addr = destip;
  addr.sin_family = AF_INET;
  
  if (connect(sock, (struct sockaddr *)&addr, sizeof(struct sockaddr)) == -1) {
      log (WARN, "connect(), ret = -1. %s\n", strerror(errno));
      return 0;
  }
  
  int len = sizeof(struct sockaddr);
  if (getsockname(sock, (struct sockaddr *)&name, (socklen_t *)&len) == -1) {
      log (WARN, "getsockname(), ret = -1. %s\n", strerror(errno));
      return 0;
  }
  
  //printf("getRoute(): src_addr = %s\n", inet_ntoa(name.sin_addr));
  
  close(sock);
  return name.sin_addr.s_addr;
}

/*
void 
NetUtil::send_tcp_packet (int sockd,
                          uint32 saddr, uint32 daddr, uint16 sport, uint16 dport,
                          uint32 seq, uint32 ack, char f_syn, char f_ack, char f_rst,
                          char* payload, int payload_len)
                          */
void 
NetUtil::send_tcp_packet (int sockd,
                          uint32 saddr, uint32 daddr, uint16 sport, uint16 dport,
                          uint32 seq, uint32 ack, char f_syn, char f_ack, char f_rst,
                          uint8 ttl)
{
    char* payload = NULL; 
    int payload_len = 0; //these two used to be arguments

    ASSERT (payload_len >= 0, "payload_len < 0");
    
    int packet_len = sizeof(struct _iphdr) + sizeof(struct _tcphdr) + payload_len;
    char* packet = (char*) malloc (packet_len);
    if (packet == NULL) throw MyException (str_log(ERROR, "malloc"));
    
    //log (DUMP, "%s:%u -> %s:%u seq=%u ack=%u payload_len=%d packet_len=%d", 
    //ip_uint32_to_str(saddr).c_str(), sport, ip_uint32_to_str(daddr).c_str(), dport,
    //seq, ack, payload_len, packet_len);
    
    memset (packet, 0, packet_len);
    
    if (payload_len > 0 && payload != NULL)
        memcpy (packet + sizeof(struct _iphdr) + sizeof(struct _tcphdr), payload, payload_len);
    
    ip_gen (packet, IPPROTO_TCP, saddr, daddr, htons(getpid()), packet_len, ttl);
    
    struct _tcphdr *tcp;
    tcp = (struct _tcphdr *)(packet + sizeof(struct _iphdr));
    
    tcp_gen ((char *)tcp, sport, dport, seq, ack, f_syn, f_ack, f_rst);
    tcp->check = trans_check(IPPROTO_TCP, (char *)tcp, sizeof(struct _tcphdr) + payload_len, saddr, daddr);
    
    struct sockaddr_in mysocket;
    memset(&mysocket, 0, sizeof(mysocket));
    //printf ("Not memset\n");
    mysocket.sin_family = AF_INET;
    mysocket.sin_port = htons(dport);
    mysocket.sin_addr.s_addr = daddr;
    
    //dumpRawData (DUMP, (uint8*)packet, packet_len);
    
    if(sendto(sockd, packet, packet_len, 0x0,
              (struct sockaddr *)&mysocket, sizeof(mysocket)) != packet_len) 
    {
        free (packet);
        throw MyException(str_log(ERROR, "sendto: %s", strerror(errno)));
    }
    free (packet);
}


void 
NetUtil::
send_udp_packet (int sockd,
                 uint32 saddr, uint32 daddr, uint16 sport, uint16 dport,
                 uint16 ip_id, uint8 ttl)
{
    int packet_len = sizeof(struct _iphdr) + sizeof(struct _udphdr);
    char* packet = (char*) malloc (packet_len);
    if (packet == NULL) throw MyException (str_log(ERROR, "malloc"));
    
    memset (packet, 0, packet_len);
    
    //if (payload_len > 0 && payload != NULL)
    //    memcpy (packet + sizeof(struct _iphdr) + sizeof(struct _tcphdr), payload, payload_len);
    //printf ("ip_id = %u\n", ip_id);

    ip_gen (packet, IPPROTO_UDP, saddr, daddr, ip_id, packet_len, ttl);
    
    struct _udphdr *udp;
    udp = (struct _udphdr *)(packet + sizeof(struct _iphdr));
    
    udp->source = htons(sport);
    udp->dest = htons(dport);
    udp->len = htons(sizeof(struct _udphdr));
    udp->check = trans_check(IPPROTO_UDP, (char *)udp, sizeof(struct _udphdr) /*+ payload_len*/, saddr, daddr);
    
    struct sockaddr_in mysocket;
    memset(&mysocket, 0, sizeof(mysocket));
    mysocket.sin_family = AF_INET;
    mysocket.sin_port = htons(dport);
    mysocket.sin_addr.s_addr = daddr;
    
    //dumpRawData (DUMP, (uint8*)packet, packet_len);
    
    if(sendto(sockd, packet, packet_len, 0x0,
              (struct sockaddr *)&mysocket, sizeof(mysocket)) != packet_len) 
    {
        free (packet);
        throw MyException(str_log(ERROR, "sendto: %s", strerror(errno)));
    }
    free (packet);
}

void
NetUtil::
send_icmp_echoreq_packet (int sockd, uint32 saddr, uint32 daddr, uint16 icmp_id, uint16 icmp_seq, uint8 ttl)
{
    int packet_len = sizeof(struct _iphdr) + sizeof(struct _icmphdr);
    char* packet = (char*) malloc (packet_len);
    if (packet == NULL) throw MyException (str_log(ERROR, "malloc"));
    
    memset (packet, 0, packet_len);
    
    ip_gen (packet, IPPROTO_ICMP, saddr, daddr, htons(getpid()), packet_len, ttl);
    
    struct _icmphdr *icmp;
    icmp = (struct _icmphdr *)(packet + sizeof(struct _iphdr));
    
    icmp->type = _ICMP_TYPE_ECHO_REQUEST;
    icmp->code = 0;
    icmp->un.echo.id = htons (icmp_id);
    icmp->un.echo.sequence = htons (icmp_seq);
    icmp->checksum = 0;
    icmp->checksum = in_cksum2 ((unsigned short*)icmp, sizeof (struct _icmphdr));
    
    struct sockaddr_in mysocket;
    memset(&mysocket, 0, sizeof(mysocket));
    mysocket.sin_family = AF_INET;
    mysocket.sin_port = htons(0);
    mysocket.sin_addr.s_addr = daddr;
    
    //dumpRawData (DUMP, (uint8*)packet, packet_len);
    
    if(sendto(sockd, packet, packet_len, 0x0,
              (struct sockaddr *)&mysocket, sizeof(mysocket)) != packet_len) 
    {
        free (packet);
        throw MyException(str_log(ERROR, "sendto: %s", strerror(errno)));
    }
    free (packet);    
}
