/*************************************************************/
/* iTraceroute (part of iSPY) Project                        */
/* Copyright (c) 2008, Zheng Zhang, Purdue University        */
/*************************************************************/

// PP_TcpAck implements TCP ACK/RST probing.

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>

#include <iostream>
#include <sstream>
#include <string>
using namespace std;

#include "NetUtil.h"
#include "PP_TcpAck.h"


void
PP_TcpAck::
create_raw_socket (int *sockd)
{
    int on = 1;
    
    int proto = IPPROTO_TCP;
#if defined __OS_CYGWIN__ || defined __OS_BSD__
    proto = IPPROTO_IP;
#elif defined __OS_LINUX__
    ;
#else
#error "NO OS"
#endif

    if((*sockd = socket(PF_INET, SOCK_RAW, proto)) < 0)  {//was IPPROTO_TCP
        throw MyException (str_log(ERROR, "socket"));
    }
    
    if(setsockopt(*sockd, IPPROTO_IP, IP_HDRINCL, (char *)&on, sizeof(on)) < 0)  {
        throw MyException (str_log(ERROR, "setsockopt"));
    }
}


PP_TcpAck::PP_TcpAck (uint32 destIP, uint16 destPort)
{
    log (INFO, "create PP_TcpAck");

    destIP_ = destIP;
    destPort_ = destPort;
    srcIP_ = destIP_ ? NetUtil::getRoute (destIP) : 0;
    srcPort_ = 0x8000 + getpid();
    
    if (destIP_) create_raw_socket (&sockd_);

    struct in_addr tmp1, tmp2;
    tmp1.s_addr = srcIP_; tmp2.s_addr = destIP_;

    log (INFO, "%s:%u -> %s:%u", string(inet_ntoa (tmp1)).c_str(), srcPort_,
         string(inet_ntoa (tmp2)).c_str(), destPort_);
}


PP_TcpAck::~PP_TcpAck ()
{
    if (destIP_) close (sockd_);
}


void
PP_TcpAck::
send (unsigned short seq)
{
    ASSERT (destIP_ != 0, "Cannot use send() to dummy probe (destIP==0)");
    log (INFO, "send probe seq = %u", seq);
    
    uint32 tcp_seq = 0;
    uint32 tcp_ack = seq;
    
    NetUtil::send_tcp_packet (sockd_, srcIP_, destIP_, srcPort_, destPort_, tcp_seq, tcp_ack, 
                              0, 1, 0); //SYN=0,ACK=1,RST=0
}


bool
PP_TcpAck::
parse (char* rawIpPacket, int packet_len, uint16* seq)
{
    struct _iphdr* ip1;
    struct _tcphdr* tcp1;
    
    if (packet_len < sizeof(struct _iphdr) + sizeof(struct _tcphdr)) {
        log (DUMP, "Bad packet. Too short.");
        return false;
    }
    ip1 = (struct _iphdr *)(rawIpPacket);
    tcp1 = (struct _tcphdr *)(rawIpPacket + sizeof(struct _iphdr));
    
    log (DUMP, "%s", NetUtil::print_ip(ip1).c_str());
    
    if (ip1->protocol != IPPROTO_TCP) {
        log (DUMP, "Bad protocol. %d. Expect TCP.", ip1->protocol);
        return false;
    }
    
    log (DUMP, NetUtil::print_tcp (tcp1).c_str());
    
    if (tcp1->rst && ntohs(tcp1->dest) == srcPort_) {
        log (INFO, "Got valid reply");
        *seq = (unsigned short)ntohl(tcp1->seq);
        return true;
    }
    
    return false;
}


string PP_TcpAck::info_str (void)
{
    ostringstream oss (ostringstream::out);

    oss << NetUtil::ip_uint32_to_str(srcIP_) << ":" << srcPort_
        << "->" << NetUtil::ip_uint32_to_str(destIP_) << ":" << destPort_
        << " tcp_ack";

    return oss.str();
}


