/*************************************************************/
/* iTraceroute (part of iSPY) Project                        */
/* Copyright (c) 2008, Zheng Zhang, Purdue University        */
/*************************************************************/

// PP_ttlTcp implements TCP TTL-limited probing.

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>

#include <iostream>
#include <sstream>
#include <string>
using namespace std;

#include "NetUtil.h"
#include "PP_ttlTcp.h"


void
PP_ttlTcp::
create_raw_socket (int *sockd)
{
    int on = 1;
    
    int proto = IPPROTO_TCP;
#if defined __OS_CYGWIN__ || defined __OS_BSD__
    proto = IPPROTO_IP;
#elif defined __OS_LINUX__
    ;
#else
#error "NO OS"
#endif

    if((*sockd = socket(PF_INET, SOCK_RAW, proto)) < 0)  {//was IPPROTO_TCP
        throw MyException (str_log(ERROR, "socket"));
    }
    
    if(setsockopt(*sockd, IPPROTO_IP, IP_HDRINCL, (char *)&on, sizeof(on)) < 0)  {
        throw MyException (str_log(ERROR, "setsockopt"));
    }
}


PP_ttlTcp::PP_ttlTcp (uint32 destIP, uint16 destPort, uint32 srcIP, uint16 srcPort)
{
    log (INFO, "create PP_ttlTcp");

    destIP_ = destIP;
    destPort_ = destPort;
    srcIP_ = srcIP ? srcIP : (destIP_ ? NetUtil::getRoute (destIP) : 0);
    srcPort_ = srcPort ? srcPort : 0x8000 + getpid();
    
    if (destIP_) create_raw_socket (&sockd_);

    struct in_addr tmp1, tmp2;
    tmp1.s_addr = srcIP_; tmp2.s_addr = destIP_;

    log (INFO, "%s:%u -> %s:%u", string(inet_ntoa (tmp1)).c_str(), srcPort_,
         string(inet_ntoa (tmp2)).c_str(), destPort_);
}


PP_ttlTcp::~PP_ttlTcp ()
{
    if (destIP_) close (sockd_);
}


void
PP_ttlTcp::
send (unsigned short seq, uint8 ttl)
{
    ASSERT (destIP_ != 0, "Cannot use send() to dummy probe (destIP==0)");
    log (INFO, "send probe seq = %u, ttl = %u", seq, ttl);
    
    uint32 tcp_seq = seq;
    uint32 tcp_ack = seq;
    
    NetUtil::send_tcp_packet (sockd_, srcIP_, destIP_, srcPort_, destPort_, tcp_seq, tcp_ack, 
                              0, 1, 0, ttl); //SYN=0,ACK=1,RST=0
}


bool
PP_ttlTcp::
parse (char* rawIpPacket, int packet_len, uint16* seq, TTL_Reply_Detail *reply_detail)
{
    struct _iphdr *ip1;

    if (packet_len < sizeof(struct _iphdr)) {
        log (DUMP, "Bad packet. Bad size.");
        return false;
    }

    ip1 = (struct _iphdr *)(rawIpPacket);

    if (ip1->protocol == IPPROTO_ICMP)
        return parse_icmp (rawIpPacket, packet_len, seq, reply_detail);
    else if (ip1->protocol == IPPROTO_TCP)
        return parse_tcp (rawIpPacket, packet_len, seq, reply_detail);
    else {
        log (DUMP, "Bad packet. Bad protocol %d.", ip1->protocol);
        return false;
    }
}


bool
PP_ttlTcp::
parse_icmp (char* rawIpPacket, int packet_len, uint16* seq, TTL_Reply_Detail *reply_detail)
{
    struct _iphdr *ip1, *ip2;
    struct _icmphdr* icmp1;
    struct _tcphdr* tcp1;

    //ICMP protocol specifies that
    //The internet header plus the first 8 bytes of the original datagram's data is returned to the sender
    if (packet_len < 2 * sizeof(struct _iphdr) + sizeof(struct _icmphdr) + 8) {
        log (DUMP, "Bad packet. Bad size.");
        return false;
    }
    //Implies that it's safe to refer to tcp1->seq
    
    ip1 = (struct _iphdr *)(rawIpPacket);
    icmp1 = (struct _icmphdr *)(rawIpPacket + sizeof(struct _iphdr));
    ip2 = (struct _iphdr *)(rawIpPacket + sizeof(struct _iphdr) + sizeof(struct _icmphdr));
    tcp1 = (struct _tcphdr *)(rawIpPacket + sizeof(struct _iphdr) + sizeof(struct _icmphdr) + sizeof(struct _iphdr));

    log (DUMP, "%s", NetUtil::print_ip(ip1).c_str());

    if (ip1->protocol != IPPROTO_ICMP || ip2->protocol != IPPROTO_TCP) {
        log (DUMP, "Bad protocol. Expect ICMP.");
        return false;
    }

    log (DUMP, NetUtil::print_icmp (icmp1).c_str());

    if (ntohs(tcp1->source) == srcPort_ && ntohs(tcp1->dest) == destPort_ &&
        (destIP_ == 0 /*DummyProbe, Don't care*/|| ip2->daddr == destIP_))
    {
        log (INFO, "Got valid reply");
        *seq = (unsigned short)ntohl(tcp1->seq);
        if (reply_detail != NULL) {
            TTL_Reply_Detail::Logic_Code lc;
            string nc;
            reply_detail_code_icmp (*icmp1, &lc, &nc);

            reply_detail->ttl = ip1->ttl;
            reply_detail->from_ip = ip1->saddr;
            reply_detail->logic_code = lc;
            reply_detail->net_code = nc;
        }
        return true;
    }
    else {
        log (DUMP, "Bad src/dst port or original destination");
//        log (DUMP, "Bad src/dst port or original destination, %u:%u  %u:%u %s %s", 
//            ntohs(tcp1->source), ntohs(tcp1->dest), 
//            srcPort_, destPort_,
//            NetUtil::ip_uint32_to_str(ip2->daddr).c_str(), 
//            NetUtil::ip_uint32_to_str(destIP_).c_str());
        return false;
    }
}


void
PP_ttlTcp::
reply_detail_code_icmp (const struct  _icmphdr & icmp, TTL_Reply_Detail::Logic_Code* lc, string* nc)
{
    if (icmp.type == 11) /* Time Exceeded                */
    {
        *lc = TTL_Reply_Detail::ON_THE_WAY;
        *nc = "";
    }
    else if (icmp.type == 3) /* Destination Unreachable      */
    {
        switch (icmp.code) {
            case 0: /* Network Unreachable          */
                *lc = TTL_Reply_Detail::UNEXPECTED;
                *nc = "!N";
                break;
            case 1: /* Host Unreachable             */
                *lc = TTL_Reply_Detail::UNEXPECTED;
                *nc = "!H";                
            case 2:  /* Protocol Unreachable         */
            case 3: /* Port Unreachable             */
                *lc = TTL_Reply_Detail::DEST_REACHED;
                *nc = "!P";
                break;
            case 4: //ICMP_SOURCE_QUENCH
                *lc = TTL_Reply_Detail::UNEXPECTED;
                *nc = "!Q";
                break;
            case 13: //ICMP_PKT_FILTERED
                *lc = TTL_Reply_Detail::UNEXPECTED;
                *nc = "!A";
                break;
            default:
                *lc = TTL_Reply_Detail::UNEXPECTED;
                *nc = "!?";
                break;
        }
    }
    else {
        *lc = TTL_Reply_Detail::UNEXPECTED;
        *nc = "!?";
    }
}


bool
PP_ttlTcp::
parse_tcp (char* rawIpPacket, int packet_len, uint16* seq, TTL_Reply_Detail *reply_detail)
{
    struct _iphdr* ip1;
    struct _tcphdr* tcp1;
    
    if (packet_len < sizeof(struct _iphdr) + sizeof(struct _tcphdr)) {
        log (DUMP, "Bad packet. Too short.");
        return false;
    }
    ip1 = (struct _iphdr *)(rawIpPacket);
    tcp1 = (struct _tcphdr *)(rawIpPacket + sizeof(struct _iphdr));
    
    log (DUMP, "%s", NetUtil::print_ip(ip1).c_str());
    
    if (ip1->protocol != IPPROTO_TCP) {
        log (DUMP, "Bad protocol. %d. Expect TCP.", ip1->protocol);
        return false;
    }
    
    log (DUMP, NetUtil::print_tcp (tcp1).c_str());
    
    if (tcp1->rst && ntohs(tcp1->dest) == srcPort_) {
        log (INFO, "Got valid reply");
        *seq = (unsigned short)ntohl(tcp1->seq);
        if (reply_detail != NULL) {
            reply_detail->ttl = ip1->ttl;
            reply_detail->from_ip = ip1->saddr;
            reply_detail->logic_code = TTL_Reply_Detail::DEST_REACHED;
            reply_detail->net_code = "!R"; //"reset"
        }
        return true;
    }
    
    return false;
}


string PP_ttlTcp::info_str (void)
{
    ostringstream oss (ostringstream::out);

    oss << NetUtil::ip_uint32_to_str(srcIP_) << ":" << srcPort_
        << "->" << NetUtil::ip_uint32_to_str(destIP_) << ":" << destPort_
        << " tcp_ack";

    return oss.str();
}



