/*************************************************************/
/* iTraceroute (part of iSPY) Project                        */
/* Copyright (c) 2008, Zheng Zhang, Purdue University        */
/*************************************************************/

// IncomPktMgr (Incoming Packet Manager) class suite allows multiple
// threads (pthread) receive incoming packets that arrive at a single
// shared socket.

#ifndef __PACKET_CATCHER_H__
#define __PACKET_CATCHER_H__

#include <sys/time.h>
#include <pthread.h>
#include <map>
#include <list>
#include <vector>
using namespace std;
#include "Common.h"

// IncomPktMgr (Incoming Packet Manager) class suite allows multiple
// threads (pthread) receive incoming packets that arrive at a single
// shared socket. IncomPktMgr provides blocking receive semantic (via
// IncomingPktMgr_Client), similar to posix recv() to these threads,
// making programming in these threads very intuitive.  An alternative
// to IncomPktMgr is to have individual threads express interest and
// provide call-back function which will be called once the shared
// socket has packets. This event-driven style breaks the code
// readability on the thread side.
// 
// What's the scenario when multiple threads waiting on a shared
// socket?
// 
// This scenario occurs in multi-threaded concurrent
// traceroute/ping. Multiple traceroute/ping threads, each of which
// works on a different probe destination, need to receive packets
// that arrive at a single shared raw socket. Why these threads don't
// use different raw socket? The reason is performance. Raw socket
// receives all packets received by the host. So if each thread use a
// different raw socket, a single packet will wake up all threads,
// however only one thread, the true receiver, actually is interested.
// 
// A IncomingPktMgr typically is linked with many
// IncomingPktMgr_Client's, each of which acts an agent of a caller
// thread, and is associated with a IncomPktMgr_Rule, which is
// responsible which of the IncomingPktMgr_Client's an incoming packet
// should be sent to.  IncomingPktMgr_Rule is the only class in this
// suite that has application knowledge (i.e. understand the packet
// content), and will be overrided.
// 
// IncomingPktMgr_Client can buffer packets. In case the caller
// threads doesn't fetch packets using the member function recv(),
// multiple packets are stored in the buffer.


struct IncomPktMgr_Packet
{
    char* pkt;
    int pkt_len;
    struct timeval recv_tm;
};

class IncomPktMgr_Client
{
public:
    IncomPktMgr_Client (void);
    ~IncomPktMgr_Client ();

// Blocking recv with timeout. Currently, a non-zero timeout must be
// specified.
    bool recv (char* pkt, int* pkt_len, struct timeval* recv_tm, int timeout_msec);
private:
    list<IncomPktMgr_Packet> pkt_buffer;
    pthread_mutex_t lock_;
    pthread_cond_t cond_wait_;
    
    friend class IncomPktMgr;
};

class IncomPktMgr_Rule
{
public:

// Map an incoming packet to a client. Return value -1 indicates that
// the rule can't decide and the next rule should be
// evaluated. Non-negative return value identifies a client.
    virtual int pkt_To_ClientId (char* rawIpPkt, int pkt_len) = 0;
};

#define MAX_PC_SUBS 256

class IncomPktMgr
{
public:

// Listen on a single protocol (raw socket). Can be IPPROTO_TCP, _UDP,
// _ICMP.
    IncomPktMgr (int protocol);

// Listen on multiple protocols (raw socket).
    IncomPktMgr (int* protocols, int nprotocol);

    ~IncomPktMgr ();

// Support multiple IncomPktMgr_Rule's. Typically one rule is
// sufficient.  Rule added first is evaluated first when receiving a
// packet. Only if the first rule returns -1, the second rule is
// evaluated, and so on.
    void append_Rule (IncomPktMgr_Rule* new_rule);

// Start the IncomingPktMgr. It's asynchronous.
    void asyn_Run (void);

    void stop (void);
    void addClient (IncomPktMgr_Client* clnt, int id);
    void removeClient (int id);

private:
    void init (int* protocols, int nprotocol);
    friend void* _run_thread (void*);
    void run_Thread (void);
    void handle_Packet (char* packet, int packlen, struct timeval tm);
    
    list<IncomPktMgr_Rule*> rule_chain;
    vector<int> protocols_;
    vector<int> sockds_;
    pthread_t thread_;
    pthread_mutex_t lock_;                     //lock the r/w of stop_thread_
    bool stop_thread_;
    IncomPktMgr_Client* subs_[MAX_PC_SUBS];
};

// Simple rule: map whatever packet to the client 0. It is used in
// PacketDMain.cc.
class IncomPktMgr_Rule_Default : public IncomPktMgr_Rule
{
public:
     int pkt_To_ClientId (char* rawIpPkt, int pkt_len) { return 0; }
};

// The rule for wtrace traceroute/ping.  The sequence number (uint16)
// of probes has a thread part and a intra-thread part.  The thread
// part is the thread index (starts from 0); the intra-thread part
// (the width is intra_thrd_bits) is the within-thread sequence
// number. IncomPktMgr_Rule_MT parses the incoming packets using a
// "dummyProbe", decodes the sequence numbers and maps packets to
// clients.
template <class ProbePacket>
class IncomPktMgr_Rule_MT : public IncomPktMgr_Rule
{
public:
    IncomPktMgr_Rule_MT (ProbePacket* dummyProbe, int intra_thrd_bits);
     int pkt_To_ClientId (char* rawIpPkt, int pkt_len);
private:
    ProbePacket* dummyProbe_;
    int intra_thrd_bits_;
};


template <class ProbePacket>
IncomPktMgr_Rule_MT<ProbePacket>::IncomPktMgr_Rule_MT (ProbePacket* dummyProbe, int intra_thrd_bits)
{
    dummyProbe_ = dummyProbe;
    intra_thrd_bits_ = intra_thrd_bits;
}


template <class ProbePacket>
int IncomPktMgr_Rule_MT<ProbePacket>::pkt_To_ClientId (char* packet, int packlen)
{
    uint16 seq;

    log (INFO, "decode_Packet, packet = %x..., packlen = %d", packet, packlen);

    ASSERT (dummyProbe_ != NULL, "");
    bool ret = dummyProbe_->parse(packet, packlen, &seq);
    if (!ret) return -1;
    
    unsigned short thrd_id = seq >> intra_thrd_bits_;
    //printf ("shift %d\n", intra_thrd_bits_);
    //printf ("seq = %u, thrd_id = %u\n", seq, thrd_id);

    return thrd_id;
}

#endif                                            //__PACKET_CATCHER_H__
